package semafori;
/**
 * Classe principale che gestisce tutte le variabili condivise
 *
 * @author Boschi Alessandro
 * @author Fabrizi Andrea
 * @author Ronchi Alessandro
 * @author Scozzoli Angelo
 */


public class Semaforo_asc
{
    /**
     * costatnti utilizzate da tutte le funzioni del programma
     * la laro scelta  fondata esclusivamnte sulle operazioni matematiche
     * che serviranno per generalizzare il codice
     */

    public final int direz_su = 3;
    public final int direz_giu = 1;

    /*
     * Dichiarazione dei semafori globali.
     * Le dimensioni dei vettori vengono definite dal costruttore della classe
     * principale "Semaforo_asc":
     */


	/**
	 * piani_coda[direzione][numero del piano]
	 */
    public Semaforo piani_coda[][];
	/**
	 * uscita_coda[id dell'ascensore][piano di destinazione]
	 */
    public Semaforo uscita_coda[][];
    public Semaforo asc_fermi_coda;
	/**
	 * piani_mutex[piano]
     *		viene usato dagli utenti per avere i diritti di acesso alle variabili
     *		condivise che specificano quante persone sono ferme ad un piano
	 */
    public Semaforo piani_mutex[];
	/**
	 * asc_mutex[ascensore + 1]
     *		viene usato dagli ascensori e dali utenti per l'accesso a variabili
     *		condivise come il contatore degli ascensori fermi
	 */
    public Semaforo asc_mutex[];
	/**
	 * * asc_piani_mutex[direzione][piano]
     *		viene usato per impedire l'apertura contemporanea di due ascensori
     *		nello stesso piano e nella stessa direzione, V() multiple su uno
     *		stesso utente
	 */
    public Semaforo asc_piani_mutex[][];
	/**
	 *  * disegna_asc[id dell'ascensore]
     *		viene usato per sincronizzare lo spostamento dell'ascensore con l'applet
	 */
    public Semaforo disegna_asc[];

    /**
     * Dichiarazione dei contattori globali delle code ai semafori e delle variabili
     * caratterizzanti lo stato e la pozizione degli ascensori:
     */

    /**
     * piani_count[direzione][piano]
     */

    public int piani_count[][];
	/**
	 * uscita_count[id ascensore][piano di uscita]
	 */
    public int uscita_count[][];
    public int pianoascensore[];
	/**
	 * asc_open[id ascensore]
	 */
    public boolean asc_open[];
	/**
	 * postiascensore[id ascensore]
	 */
    public int postiascensore[];
	/**
	 * asc_direz[id ascensore]
	 */
    public int asc_direz[];
    public int n_piani;
    public int n_ascensori;
    public int asc_fermi_count;
	/**
	 * n_graph[id ascensore]
     *		serve per assicurari che le gli utenti non facciano mai partirel'ascensore
     *		prima che l'applet habbia finito di disegnare il loro movimento
     */
    public int n_graph[];

    /**
     * Costruttore della classe principale.
     * di questa classe deve essere istanziato un solo ogetto che poi verr passato
     * alle altre classi
     * @param na numero di ascensori
     * @param np numero di piani
     */

    public Semaforo_asc(int np, int na)
    {
	int i;
	int j;
	n_piani=np;
	n_ascensori=na;
	asc_fermi_count=0;
	asc_fermi_coda=new Semaforo(0);

	// predichiarazioni di lunghezza dei vettori

	asc_mutex = new Semaforo[na+1];
	piani_coda = new Semaforo[4][np];
	piani_count = new int [4][np];
	piani_mutex = new Semaforo[np];
	uscita_coda = new Semaforo[na][np];
	asc_piani_mutex = new Semaforo[4][np];
	uscita_count = new int [na][np];
	pianoascensore = new int[na];
	asc_open = new boolean[na];
	postiascensore = new int[na];
	asc_direz= new int[na];
	disegna_asc= new Semaforo[na];
	n_graph = new int[na];
	//end predichiarazione

	asc_mutex[na]=new Semaforo(1);
	// disegna=new Semaforo(0);
	for(i=0;i<np;i++)
	    {
		piani_coda[direz_su][i]=new Semaforo(0);
		piani_coda[direz_giu][i]=new Semaforo(0);
		piani_count[direz_su][i]=0;
		piani_count[direz_giu][i]=0;
		piani_mutex[i]=new Semaforo(1);
		asc_piani_mutex[direz_su][i]=new Semaforo(1);
		asc_piani_mutex[direz_giu][i]=new Semaforo(1);
		for (j=0;j<na;j++)
		    {
			uscita_coda[j][i]=new Semaforo(0);
			uscita_count[j][i]=0;
		    }
	    }
	for(j=0;j<na;j++)
	    {
		asc_mutex[j]=new Semaforo(0);
		pianoascensore[j]=0;
		asc_open[j]=false;
		asc_direz[j]=direz_su;
		disegna_asc[j]=new Semaforo(0);
		n_graph[j]=0;
	    }
    }// end costruttore

    /**
     * Funzione usata dagli ascensori prima di mettersi in riposo, per sapere
     * se qualche utente  entrato in coda nela frattempo.
     * @return true se non c' nessuno, false se c' qalcuno.
     */

    public boolean no_one()
    {
	int i;
	i=0;
	while (i<n_piani)
	    {
		if ((piani_count[direz_giu][i]!=0)||(piani_count[direz_giu][i]!=0))
		    return false;
		i++;
	    }
	return true;
    }

    /**
     * Funzioni usate dalla classe ascensore
     */

    public void asc_init()
    {
	asc_mutex[n_ascensori].P();
	if (no_one())
	    {
		asc_fermi_count++;
		asc_mutex[n_ascensori].V();
		asc_fermi_coda.P();
	    }
	else
	    {
		asc_mutex[n_ascensori].V();
	    }
    }

    /**
     * Procedura per il movimento dell'ascensore
     * L'ascensore si muove di piano in piano guardando se c' qualcuno da scaricare
     * e se c' nessuno che vada nella sua direzione da caricare.
     * La procedura risveglia solamente il primo utente in ogni piano, sar poi
     * lui che risveglier quelli successivi.
     * Sono infine sempre gli utenti che risvegliaranno l'ascensore una volta finito
     * @see usr_enter(persona usr)
     * @param piano_dest piano verso cui si deve muovere l'ascensore
     * @param asc la classe stessa dell'ascensore che si muove
     */

    public void asc_muovi(int piano_dest, ascensore asc)
    {
	int i;
	while((piano_dest!=(pianoascensore[asc.id]))&&
	      ((piani_count[direz_su][piano_dest]!=0 )||
	       (piani_count[direz_giu][piano_dest]!=0)||
	       (uscita_count[asc.id][piano_dest]!=0  )))
	    {
		while(uscita_count[asc.id][pianoascensore[asc.id]]!=0)
		    {
			uscita_coda[asc.id][pianoascensore[asc.id]].V();
			postiascensore[asc.id]++;
			uscita_count[asc.id][pianoascensore[asc.id]]--;
		    }
                while((piani_count[asc_direz[asc.id]][pianoascensore[asc.id]]!=0)&&(postiascensore[asc.id]!=0))
                {
		    asc_piani_mutex[asc_direz[asc.id]][pianoascensore[asc.id]].P();
		    if ((piani_count[asc_direz[asc.id]][pianoascensore[asc.id]]!=0)&&
		        (postiascensore[asc.id]!=0))
		        {
			    asc_open[asc.id]=true;
			    piani_coda[asc_direz[asc.id]][pianoascensore[asc.id]].V();
			    asc_mutex[asc.id].P();
		        }
		    else
		        {
			    asc_piani_mutex[asc_direz[asc.id]][pianoascensore[asc.id]].V();
                        }
                }
                pianoascensore[asc.id] = (pianoascensore[asc.id]+(asc_direz[asc.id] -2));
		disegna_asc[asc.id].P();
	    }//end while

	/**
	 * Quando l'ascensore raggiunge il piano di destinazione controlla se l'utente che
	 * l'ha chiamato ha la sua stessa direzione: se si lo carica ... se no
	 * cambia la sua direzione
	 */

	while(uscita_count[asc.id][pianoascensore[asc.id]]!=0)
	    {
		uscita_coda[asc.id][pianoascensore[asc.id]].V();
		postiascensore[asc.id]++;
		uscita_count[asc.id][pianoascensore[asc.id]]--;
	    }
	asc_piani_mutex[asc_direz[asc.id]][pianoascensore[asc.id]].P();
	if((piano_dest==pianoascensore[asc.id]) &&
	   (piani_count[asc_direz[asc.id]][pianoascensore[asc.id]]!=0))
	    {
		asc_open[asc.id]=true;
		piani_coda[asc_direz[asc.id]][pianoascensore[asc.id]].V();
		asc_mutex[asc.id].P();
	    }
	else
	    {
		asc_piani_mutex[asc_direz[asc.id]][pianoascensore[asc.id]].V();
	    }



    }//end asc_muovi

    /**
     * Funzioni usate dalla classe utente
     */

    /**
     * Procedura che inizializza l'utente.
     * La persona che viene creata, per prima cosa cosa incrementa il contatore
     * delle persone in conda, poi quarda se ci sono ascensori svegli da risvegliare
     * ed in fine si mette in sul semaforo dei piani.
     * @param usr la classe stessa dell'utente che si inizializza.
     */

    public void usr_init(persona usr)
    {
	piani_mutex[usr.piano_from].P();
	piani_count[usr.direzione][usr.piano_from]++;
	piani_mutex[usr.piano_from].V();
	asc_mutex[n_ascensori].P();
	while (asc_fermi_count!=0)
	    {
		asc_fermi_coda.V();
		asc_fermi_count--;
	    }
	asc_mutex[n_ascensori].V();
	piani_coda[usr.direzione][usr.piano_from].P();
    }//end usr_init

    /**
     * Procedura di entrata dell'utente.
     * Per ogni piano l'ascensore risveglia solo il primo utente che risveglier
     * quelli successivi se ci sono posti nell'ascensore svegliato e se c' ancora
     * qualcuno che aspetta nella coda del proprio piano e propria direzione.
     * L'ultimo sblocca la mutua esclusione sul piano, ma il thread che sblocher
     * l'ascensore  quello che avr il contatoren_graph[id ascenmsore]==0.
     * Difatti tale contatore viene incrementato dagli utenti che devono entrare,
     * ma viene decrementato quando l'applet avr finito di graficarli.
     * @see asc_muovi(int piano_dest, ascensore asc)
     * @param usr la classe stessa dell'utente che entra nell'ascensore.
     */

    public void usr_enter(persona usr)
    {
	int i;
	i=0;
	/**
	 * Questo ciclo while serve ad individuare l'ascdensore che ha svegliato l'utente
	 */
	while ((!asc_open[i])||(pianoascensore[i]!=usr.piano_from)||(asc_direz[i]!=usr.direzione))
	    i++;
	piani_mutex[usr.piano_from].P();
        usr.imagefile=("pass"+ usr.usr_img +"_"+ i +".gif");
	postiascensore[i]--;
	uscita_count[i][usr.piano_to]++;
	piani_count[usr.direzione][usr.piano_from]--;
	usr.id_asc=i;
	usr.muovi=true;
	n_graph[i]++;
	/**
	 * Gli utenti si svegliano l'un l'altro a cascata finch ci sono persone in coda
	 * e posti nell'ascensore
	 */
	if (((postiascensore[i])!=0)&&(piani_count[usr.direzione][usr.piano_from]!=0))
	    {
		piani_coda[usr.direzione][usr.piano_from].V();
		piani_mutex[usr.piano_from].V();
		usr.disegna.P();
		usr.muovi=false;
		n_graph[i]--;
	    }
	else
	    {
		asc_open[i]=false;
		asc_piani_mutex[asc_direz[i]][pianoascensore[i]].V();
		//usr.muovi=false;
		piani_mutex[usr.piano_from].V();
		usr.disegna.P();
		usr.muovi=false;
		n_graph[i]--;
	    }
	if (n_graph[i]==0)
	    asc_mutex[i].V();
	uscita_coda[i][usr.piano_to].P();
    }//end usr_enter

}
